/*	$NetBSD$	*/

/*-
 * Copyright (c) 2001 Minoura Makoto.
 * Copyright (c) 1994 Charles M. Hannum.
 * Copyright (c) 1992, 1993 Erik Forsberg.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * THIS SOFTWARE IS PROVIDED BY ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL I BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Synaptics TouchPad via PS/2 aux port.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/device.h>
#include <sys/ioctl.h>

#include <machine/bus.h>

#include <dev/ic/stpreg.h>
#include <dev/ic/stpvar.h>
#include <dev/ic/pckbcvar.h>

#include <dev/pckbc/psmreg.h>
#include <dev/wscons/wsconsio.h>
#include <dev/wscons/wsmousevar.h>

#define STP_PCKBC_CMD_IDENT		0
#define STP_PCKBC_CMD_READ_MODE		1
#define STP_PCKBC_CMD_READ_CAPS		2
#define STP_PCKBC_CMD_READ_MODEL	3
#define STP_PCKBC_CMD_READ_SERP		6
#define STP_PCKBC_CMD_READ_SERS		7
#define STP_PCKBC_CMD_READ_RES		8

struct stp_pckbc_softc {
	struct stp_softc sc_stp;

	pckbc_tag_t sc_kbctag;
	int sc_kbcslot;

	int sc_enabled;		/* input enabled? */
	int inputstate;
	u_int buttons, oldbuttons;	/* mouse button status */
	signed char dx;
};

/* autoconf methods */
static int stp_pckbc_probe __P((struct device *, struct cfdata *, void *));
static void stp_pckbc_attach __P((struct device *, struct device *, void *));

struct cfattach stp_pckbc_ca = {
	sizeof (struct stp_pckbc_softc), stp_pckbc_probe, stp_pckbc_attach,
};

/* pckbc mothods */
static void stp_pckbc_input __P((void *, int));

/* wsmouse methods */
static int stp_pckbc_enable __P((void *));
static int stp_ioctl __P((void *, u_long, caddr_t, int, struct proc *));
static void stp_pckbc_disable __P((void *));

const struct wsmouse_accessops stp_accessops = {
	stp_pckbc_enable,
	stp_ioctl,
	stp_pckbc_disable,
};

/* stp methods */
static int stp_pckbc_getmode __P((void *, u_int8_t *));
static int stp_pckbc_setmode __P((void *, u_int8_t));
static int stp_pckbc_identify __P((void *, stp_identifier *));

/* internal functions */
static int stp_pckbc_query __P((pckbc_tag_t, pckbc_slot_t, int, u_char []));


/*
 * generic information query function for Synaptics
 */ 
static int stp_pckbc_query(tag, slot, command, response)
	pckbc_tag_t tag;
	pckbc_slot_t slot;
	int command;
	u_char response[];	/* at least 3bytes */
{
	u_char		cmdseq[2];
	int i, r;

	cmdseq[0] = PMS_SET_RES;
	for (i = 6; i >= 0; i -= 2) {
		cmdseq[1] = (command >> i) & 3;
		r = pckbc_poll_cmd(tag, slot, cmdseq, 2, 0, 0, 0);
		if (r)
			return r;
	}

	cmdseq[0] = PMS_GET_SCALE;
	r = pckbc_poll_cmd(tag, slot, cmdseq, 1, 3, response, 0);
	if (r)
		return r;

	return 0;
}

/*
 * stp methods
 */
static int stp_pckbc_identify(arg, ident)
	void *arg;
	stp_identifier *ident;
{
	struct stp_pckbc_softc *sc = arg;
	pckbc_tag_t tag = sc->sc_kbctag;
	pckbc_slot_t slot = sc->sc_kbcslot;
	u_char resp[3];
	int r;

	r = stp_pckbc_query(tag, slot, STP_PCKBC_CMD_IDENT, resp);
	if (r)
		return r;
	if (resp[1] != 0x47)
		return EIO;
	ident->si_info_major = resp[2] & 0x0f;
	ident->si_info_minor = resp[0];
	ident->si_info_modelcode = resp[2] >> 4;

	r = stp_pckbc_query(tag, slot, STP_PCKBC_CMD_READ_CAPS, resp);
	if (r)
		return r;
	if (resp[1] != 0x47)
		return EIO;
	ident->si_capabilities = (resp[0] << 8) | resp[2];
	r = stp_pckbc_query(tag, slot, STP_PCKBC_CMD_READ_MODEL, resp);
	if (r)
		return r;
	ident->si_model_id = (resp[0] << 16) | (resp[1] << 8) | resp[0];

	return 0;
}

static int stp_pckbc_getmode(arg, result)
	void *arg;
	u_int8_t *result;
{
	struct stp_pckbc_softc *sc = arg;
	pckbc_tag_t tag = sc->sc_kbctag;
	pckbc_slot_t slot = sc->sc_kbcslot;
	u_char resp[3];
	int r;

	r = stp_pckbc_query(tag, slot, STP_PCKBC_CMD_READ_MODE, resp);
	if (r) {
		printf ("stp_pckbc_getmode: read_mode failed.\n");
		return r;
	}
	if (resp[0] != 0x3b || resp[1] != 0x47) {
		printf ("stp_pckbc_getmode: unexpected response.\n");
		return r;
	}

	*result = resp[2];
	return 0;
}

static int stp_pckbc_setmode(arg, mode)
	void *arg;
	u_int8_t mode;
{
	struct stp_pckbc_softc *sc = arg;
	pckbc_tag_t tag = sc->sc_kbctag;
	pckbc_slot_t slot = sc->sc_kbcslot;
	u_char cmdseq[2];
	int i, r;

	cmdseq[0] = PMS_SET_RES;
	for (i = 6; i >= 0; i -= 2) {
		cmdseq[1] = (mode >> i) & 3;
		r = pckbc_poll_cmd(tag, slot, cmdseq, 2, 0, 0, 0);
		if (r)
			return r;
	}

	cmdseq[0] = PMS_SET_SAMPLE;
	cmdseq[1] = 20;
	r = pckbc_poll_cmd(tag, slot, cmdseq, 2, 0, 0, 0);
	if (r)
		return r;

	return 0;
}

/*
 * autoconf methods
 */
static int stp_pckbc_probe(parent, match, aux)
	struct device *parent;
	struct cfdata *match;
	void *aux;
{
	struct pckbc_attach_args *pa = aux;
	u_char cmd[1], resp[3];
	int res;

	if (pa->pa_slot != PCKBC_AUX_SLOT)
		return 0;

	/* Flush any garbage. */
	pckbc_flush(pa->pa_tag, pa->pa_slot);

	/* reset the device */
	cmd[0] = PMS_RESET;
	res = pckbc_poll_cmd(pa->pa_tag, pa->pa_slot, cmd, 1, 2, resp, 1);
	if (res) {
#ifdef DEBUG
		printf("stp_pckbc_probe: reset error %d\n", res);
#endif
		return (0);
	}
	if (resp[0] != PMS_RSTDONE) {
		printf("stp_pckbc_probe: reset response 0x%x\n", resp[0]);
		return (0);
	}

	/* get type number (0 = mouse) */
	if (resp[1] != 0) {
#ifdef DEBUG
		printf("stp_pckbc_probe: type 0x%x\n", resp[1]);
#endif
		return (0);
	}

	
	res = stp_pckbc_query(pa->pa_tag, pa->pa_slot,
			      STP_PCKBC_CMD_IDENT, resp);
	if (res)
		return 0;
	if (resp[1] != 0x47)
		return 0;

	return 20;
}

static void stp_pckbc_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct stp_pckbc_softc *sc = (struct stp_pckbc_softc *) self;
	struct pckbc_attach_args *pa = aux;
	struct wsmousedev_attach_args a;
	int r;
	u_char cmd[1], resp[2];

	sc->sc_kbctag = pa->pa_tag;
	sc->sc_kbcslot = pa->pa_slot;

	/* Flush any garbage. */
	pckbc_flush(pa->pa_tag, pa->pa_slot);

	/* reset the device */
	cmd[0] = PMS_RESET;
	r = pckbc_poll_cmd(pa->pa_tag, pa->pa_slot, cmd, 1, 2, resp, 1);
#ifdef DEBUG
	if (r || resp[0] != PMS_RSTDONE || resp[1] != 0) {
		printf("\nstp_pckbc_attach: reset error\n");
		return;
	}
#endif

	sc->sc_stp.sc_identify = stp_pckbc_identify;
	sc->sc_stp.sc_getmode = stp_pckbc_getmode;
	sc->sc_stp.sc_setmode = stp_pckbc_setmode;
	stp_init(sc);

	sc->inputstate = 0;
	sc->oldbuttons = 0;

	pckbc_set_inputhandler(sc->sc_kbctag, sc->sc_kbcslot,
			       stp_pckbc_input, sc,
			       sc->sc_stp.sc_dev.dv_xname);

	a.accessops = &stp_accessops;
	a.accesscookie = sc;

	/*
	 * Attach the wsmouse, saving a handle to it.
	 */
	sc->sc_stp.sc_wsmousedev = config_found(self, &a, wsmousedevprint);

	/* no interrupts until enabled */
	cmd[0] = PMS_DEV_DISABLE;
	r = pckbc_poll_cmd(pa->pa_tag, pa->pa_slot, cmd, 1, 0, 0, 0);
	if (r)
		printf("stp_pckbc_attach: disable error\n");
	pckbc_slot_enable(sc->sc_kbctag, sc->sc_kbcslot, 0);
}

/*
 * wsmouse methods
 */
static int stp_pckbc_enable(v)
	void *v;
{
	struct stp_pckbc_softc *sc = v;
	u_char cmd[1];
	int res;

	if (sc->sc_enabled)
		return EBUSY;

	sc->sc_enabled = 1;
	sc->inputstate = 0;
	sc->oldbuttons = 0;

	pckbc_slot_enable(sc->sc_kbctag, sc->sc_kbcslot, 1);

	cmd[0] = PMS_DEV_ENABLE;
	res = pckbc_enqueue_cmd(sc->sc_kbctag, sc->sc_kbcslot,
				cmd, 1, 0, 1, 0);
	if (res)
		printf("stp_pckbc_enable: command error\n");

	return 0;
}

static void stp_pckbc_disable(v)
	void *v;
{
	struct stp_pckbc_softc *sc = v;
	u_char cmd[1];
	int res;

	cmd[0] = PMS_DEV_DISABLE;
	res = pckbc_enqueue_cmd(sc->sc_kbctag, sc->sc_kbcslot,
				cmd, 1, 0, 1, 0);
	if (res)
		printf("stp_pckbc_disable: command error\n");

	pckbc_slot_enable(sc->sc_kbctag, sc->sc_kbcslot, 0);

	sc->sc_enabled = 0;
}

static int stp_ioctl(v, cmd, data, flag, p)
	void *v;
	u_long cmd;
	caddr_t data;
	int flag;
	struct proc *p;
{
	struct stp_pckbc_softc *sc = v;
	u_char kbcmd[2];
	int i;

	switch (cmd) {
	case WSMOUSEIO_GTYPE:
		*(u_int *)data = WSMOUSE_TYPE_TPANEL;
		break;
		
	case WSMOUSEIO_SRES:
		i = (*(u_int *)data - 12) / 25;
		
		if (i < 0)
			i = 0;
			
		if (i > 3)
			i = 3;

		kbcmd[0] = PMS_SET_RES;
		kbcmd[1] = i;			
		i = pckbc_enqueue_cmd(sc->sc_kbctag, sc->sc_kbcslot,
				      kbcmd, 2, 0, 1, 0);
		
		if (i)
			printf("stp_pckbc_ioctl: SET_RES command error\n");
		break;
		
	default:
		return (-1);
	}
	return (0);
}

/*
 * pckbc interrupt routine
 */

/* Masks for the first byte of a packet */
#define PS2LBUTMASK 0x01
#define PS2RBUTMASK 0x02
#define PS2MBUTMASK 0x04

static void stp_pckbc_input(vsc, data)
	void *vsc;
	int data;
{
	struct stp_pckbc_softc *sc = vsc;
	signed char dy;
	u_int changed;

	if (!sc->sc_enabled) {
		/* Interrupts are not expected.  Discard the byte. */
		return;
	}

	switch (sc->inputstate) {

	case 0:
		if ((data & 0xc0) == 0) { /* no ovfl, bit 3 == 1 too? */
			sc->buttons = ((data & PS2LBUTMASK) ? 0x1 : 0) |
			    ((data & PS2MBUTMASK) ? 0x2 : 0) |
			    ((data & PS2RBUTMASK) ? 0x4 : 0);
			++sc->inputstate;
		}
		break;

	case 1:
		sc->dx = data;
		/* Bounding at -127 avoids a bug in XFree86. */
		sc->dx = (sc->dx == -128) ? -127 : sc->dx;
		++sc->inputstate;
		break;

	case 2:
		dy = data;
		dy = (dy == -128) ? -127 : dy;
		sc->inputstate = 0;

		changed = (sc->buttons ^ sc->oldbuttons);
		sc->oldbuttons = sc->buttons;

		if (sc->dx || dy || changed)
			wsmouse_input(sc->sc_stp.sc_wsmousedev,
				      sc->buttons, sc->dx, dy, 0,
				      WSMOUSE_INPUT_DELTA);
		break;
	}

	return;
}
